@@ -26,6 +26,9 @@
 #ifndef LISP_FEATURE_WIN32
 #include <sys/wait.h>
 #endif
+#ifdef ZSTD_STATIC_LINKING_ONLY
+#include <zstd.h>
+#endif
 
 #include "runtime.h"
 #include "validate.h"           /* for BINDING_STACK_SIZE etc */
@@ -40,6 +43,7 @@
 #include "genesis/instance.h"
 #include "genesis/vector.h"
 #include "interr.h"             /* for lose() */
+#include "gc-assert.h"
 #include "gc.h"
 #include "pseudo-atomic.h"
 #include "interrupt.h"
@@ -86,9 +90,23 @@ static pthread_mutex_t in_gc_lock = PTHREAD_MUTEX_INITIALIZER;
 extern lispobj call_into_lisp_first_time(lispobj fun, lispobj *args, int nargs);
 #endif
 
+#ifdef ZSTD_STATIC_LINKING_ONLY
+static bool asan_cleanup_called;
+#endif
 static void
 link_thread(struct thread *th)
 {
+#ifdef ZSTD_STATIC_LINKING_ONLY
+    if (!asan_cleanup_called) {
+    /* A thread has a preallocated dcontext if and only if it is in all_threads,
+     * unless cleanup has already been called, in which case we just hope that
+     * no backtrace occurs thereafter, or else that allocating a context
+     * just-in-time (in decompress_vector) works, which it may not */
+        struct extra_thread_data *extra_data = thread_extra_data(th);
+        gc_assert(!extra_data->zstd_dcontext);
+        extra_data->zstd_dcontext = ZSTD_createDCtx();
+    }
+#endif
     if (all_threads) all_threads->prev=th;
     th->next=all_threads;
     th->prev=0;
@@ -99,6 +117,11 @@ link_thread(struct thread *th)
 static void
 unlink_thread(struct thread *th)
 {
+#ifdef ZSTD_STATIC_LINKING_ONLY
+    struct extra_thread_data *extra_data = thread_extra_data(th);
+    if (extra_data->zstd_dcontext) ZSTD_freeDCtx(extra_data->zstd_dcontext);
+    extra_data->zstd_dcontext = 0;
+#endif
     if (th->prev)
         th->prev->next = th->next;
     else
@@ -121,7 +144,7 @@ char* vm_thread_name(struct thread* th)
 #define get_thread_state(thread) \
  (int)__sync_val_compare_and_swap(&thread->state_word.state, -1, -1)
 
-#ifndef LISP_FEATURE_SB_SAFEPOINT
+#if HAVE_GC_STW_SIGNAL
 
 void
 set_thread_state(struct thread *thread,
@@ -228,6 +251,10 @@ int sb_GetTID()
 int sb_GetTID() {
     return pthread_mach_thread_np(pthread_self());
 }
+#elif defined __HAIKU__ && defined LISP_FEATURE_SB_THREAD
+int sb_GetTID() {
+    return get_pthread_thread_id(pthread_self());
+}
 #else
 #define sb_GetTID() 0
 #endif
@@ -312,21 +339,32 @@ void reset_gc_stats() { // after sb-posix:fork
 }
 #endif
 
-#ifdef ATOMIC_LOGGING
-#define THREAD_NAME_MAP_MAX 20 /* KLUDGE */
-struct {
-  pthread_t thread;
-  char *name; // strdup'ed
-} thread_name_map[THREAD_NAME_MAP_MAX];
-int thread_name_map_count;
-
-char* thread_name_from_pthread(pthread_t pointer){
-    int i;
-    for(i=0; i<thread_name_map_count; ++i)
-        if (thread_name_map[i].thread == pointer) return thread_name_map[i].name;
-    return 0;
-}
+/* This function is seemingly unused by the C runtime, but DO NOT DELETE.
+ * It's needed when the C runtime is compiled with --fsanitize=address, because the sanitizer
+ * falsely reports that all the threads running at exit have leaked their zstd_dcontext.
+ * While it's possible to make the sanitizer shut up about particular allocations,
+ * you need to know the size in order to pass it to the unpoisoning routine.
+ * The allocator of the ZSTD context object is opaque; we don't know the size.
+ * Interestingly the sanitizer does not complain about the thread structure per se,
+ * and I think that's because it does not track mmap().
+ * I wanted to register this function using atexit() but apparently that's not soon enough
+ * to solve the problem. It has to to be called by *EXIT-HOOKS* instead */
+void asan_lisp_thread_cleanup() {
+    ignore_value(mutex_acquire(&all_threads_lock));
+#if defined ADDRESS_SANITIZER && defined ZSTD_STATIC_LINKING_ONLY
+    asan_cleanup_called = 1;
+    struct thread* th;
+    for_each_thread(th) {
+        struct extra_thread_data *extra_data = thread_extra_data(th);
+        void* dctx = extra_data->zstd_dcontext;
+        if (dctx && __sync_bool_compare_and_swap(&extra_data->zstd_dcontext, dctx, 0))
+            ZSTD_freeDCtx(dctx);
+    }
 #endif
+    ignore_value(mutex_release(&all_threads_lock));
+}
+
+static void unregister_thread(struct thread *, init_thread_data *);
 
 void create_main_lisp_thread(lispobj function) {
 #ifdef LISP_FEATURE_WIN32
@@ -349,7 +387,7 @@ void create_main_lisp_thread(lispobj function) {
 #endif
     ASSOCIATE_OS_THREAD(th);
     ASSIGN_CURRENT_THREAD(th);
-#if defined THREADS_USING_GCSIGNAL && \
+#if THREADS_USING_GCSIGNAL && \
     (defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64 || defined LISP_FEATURE_ARM64 || defined LISP_FEATURE_RISCV)
     /* SIG_STOP_FOR_GC defaults to blocked on PPC? */
     unblock_gc_stop_signal();
@@ -372,6 +410,9 @@ void create_main_lisp_thread(lispobj function) {
     set_thread_stack(th->control_stack_end);
 #endif
 
+#ifdef LISP_FEATURE_PPC64
+    NIL_SYMBOL_SLOTS_START[-1] = (uword_t)gc_card_mark;
+#endif
 #ifdef COLLECT_GC_STATS
     atexit(summarize_gc_stats);
 #endif
@@ -384,14 +425,10 @@ void create_main_lisp_thread(lispobj function) {
 #else
     funcall0(function);
 #endif
-    // If we end up returning, clean up the initial thread.
+    // If we end up returning (when used as shared library), clean up the initial thread.
 #ifdef LISP_FEATURE_SB_THREAD
-    unlink_thread(th);
-#else
-    all_threads = NULL;
+    unregister_thread(th, NULL);
 #endif
-    arch_os_thread_cleanup(th);
-    ASSIGN_CURRENT_THREAD(NULL);
 }
 
 void sb_posix_after_fork() { // for use by sb-posix:fork
@@ -487,7 +524,8 @@ unregister_thread(struct thread *th,
 #endif
     gc_close_thread_regions(th, LOCK_PAGE_TABLE|CONSUME_REMAINDER);
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-    pop_gcing_safety(&scribble->safety);
+    if (scribble)
+      pop_gcing_safety(&scribble->safety);
 #else
     /* This state change serves to "acknowledge" any stop-the-world
      * signal received while the STOP_FOR_GC signal is blocked */
@@ -509,7 +547,7 @@ unregister_thread(struct thread *th,
 #ifdef LISP_FEATURE_UNIX
     os_sem_destroy(&semaphores->sprof_sem);
 #endif
-#ifndef LISP_FEATURE_SB_SAFEPOINT
+#if HAVE_GC_STW_SIGNAL
     os_sem_destroy(&semaphores->state_sem);
     os_sem_destroy(&semaphores->state_not_running_sem);
     os_sem_destroy(&semaphores->state_not_stopped_sem);
@@ -520,6 +558,14 @@ unregister_thread(struct thread *th,
     for (i = 0; i<NUM_PRIVATE_EVENTS; ++i)
         CloseHandle(thread_private_events(th,i));
 #endif
+    /* Thread structs are reused and their guard pages are not
+     * reestablished. */
+    if (!th->state_word.control_stack_guard_page_protected)
+        reset_thread_control_stack_guard_page(th);
+    if (!th->state_word.alien_stack_guard_page_protected)
+        reset_thread_alien_stack_guard_page(th);
+    if (!th->state_word.binding_stack_guard_page_protected)
+        reset_thread_binding_stack_guard_page(th);
 
     /* Undo the association of the current pthread to its `struct thread',
      * such that we can call get_sb_vm_thread() later in this
@@ -563,14 +609,6 @@ void* new_thread_trampoline(void* arg)
     struct thread_instance *lispthread = (void*)native_pointer(th->lisp_thread);
     if (lispthread->_ephemeral_p == LISP_T) th->state_word.user_thread_p = 0;
 
-#ifdef ATOMIC_LOGGING
-      char* string = strdup((char*)VECTOR(name)->data); // FIXME: no such var as 'name'
-      int index = __sync_fetch_and_add(&thread_name_map_count, 1);
-      gc_assert(index < THREAD_NAME_MAP_MAX);
-      thread_name_map[index].thread = pthread_self();
-      thread_name_map[index].name = string;
-#endif
-
     struct vector* startup_info = VECTOR(lispthread->startup_info); // 'lispthread' is pinned
     gc_assert(header_widetag(startup_info->header) == SIMPLE_VECTOR_WIDETAG);
     lispobj startfun = startup_info->data[0]; // 'startup_info' is pinned
@@ -700,7 +738,7 @@ static void attach_os_thread(init_thread_data *scribble)
     void* recycled_memory = get_recyclebin_item();
     struct thread *th = alloc_thread_struct(recycled_memory);
 
-#ifndef LISP_FEATURE_SB_SAFEPOINT
+#if HAVE_GC_STW_SIGNAL
     /* new-lisp-thread-trampoline doesn't like when the GC signal is blocked */
     /* FIXME: could be done using a single call to pthread_sigmask
        together with blocking the deferrable signals above. */
@@ -783,7 +821,7 @@ static void detach_os_thread(init_thread_data *scribble)
      *  - but STOP_FOR_GC is pending because it was in the blocked set.
      * Bad things happen unless we clear the pending GC signal.
      */
-#if !defined LISP_FEATURE_SB_SAFEPOINT
+#if HAVE_GC_STW_SIGNAL
     sigset_t pending;
     sigpending(&pending);
     if (sigismember(&pending, SIG_STOP_FOR_GC)) {
@@ -829,17 +867,7 @@ extern void funcall_alien_callback(lispobj arg1, lispobj arg2, lispobj arg0,
 /* This function's address is assigned into a static symbol's value slot,
  * so it has to look like a fixnum. lp#1991485 */
 void __attribute__((aligned(8)))
-callback_wrapper_trampoline(
-#if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
-    /* On the x86oid backends, the assembly wrapper happens to not pass
-     * in ENTER_ALIEN_CALLBACK explicitly for safepoints.  However, the
-     * platforms with precise GC are tricky enough already, and I want
-     * to minimize the read-time conditionals.  For those platforms, I'm
-     * only replacing funcall3 with callback_wrapper_trampoline while
-     * keeping the arguments unchanged. --DFL */
-    lispobj __attribute__((__unused__)) fun,
-#endif
-    lispobj arg0, lispobj arg1, lispobj arg2)
+callback_wrapper_trampoline(lispobj arg0, lispobj arg1, lispobj arg2)
 {
     struct thread* th = get_sb_vm_thread();
     if (!th) {                  /* callback invoked in non-lisp thread */
@@ -882,8 +910,8 @@ callback_wrapper_trampoline(
  *     /   ___ aligned_spaces
  *    /   /
  *  (0) (1)       (2)       (3)       (4)    (5)          (6)
- *   |   | CONTROL | BINDING |  ALIEN  |  CSP | thread     |          |
- *   |   |  STACK  |  STACK  |  STACK  | PAGE | structure  | altstack |
+ *   |   | CONTROL | BINDING |  ALIEN  | Trap | thread     |          |
+ *   |   |  STACK  |  STACK  |  STACK  | page | structure  | altstack |
  *   |...|------------------------------------------------------------|
  *          2MiB       1MiB     1MiB               (*)         (**)
  *
@@ -920,14 +948,14 @@ alloc_thread_struct(void* spaces) {
      * from failing to obtain contiguous memory. Note that the OS may have a smaller
      * alignment granularity than BACKEND_PAGE_BYTES so we may have to adjust the
      * result to make it conform to our guard page alignment requirement. */
-    bool zeroize_stack = 0;
+    bool is_recycled = 0;
     if (spaces) {
         // If reusing memory from a previously exited thread, start by removing
         // some old junk from the stack. This is imperfect since we only clear a little
         // at the top, but doing so enables diagnosing some garbage-retention issues
         // using a fine-toothed comb. It would not be possible at all to diagnose
         // if any newly started thread could refer a dead thread's heap objects.
-        zeroize_stack = 1;
+        is_recycled = 1;
     } else {
         spaces = os_alloc_gc_space(THREAD_STRUCT_CORE_SPACE_ID, MOVABLE,
                                    NULL, THREAD_STRUCT_SIZE);
@@ -958,25 +986,6 @@ alloc_thread_struct(void* spaces) {
 #endif
 
     __attribute((unused)) lispobj* tls = (lispobj*)th;
-#ifdef THREAD_T_NIL_CONSTANTS_SLOT
-    tls[THREAD_T_NIL_CONSTANTS_SLOT] = (NIL << 32) | LISP_T;
-#endif
-#ifdef LISP_FEATURE_LINKAGE_SPACE
-    tls[THREAD_LINKAGE_TABLE_SLOT] = (lispobj)linkage_space;
-    tls[THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT] = (lispobj)ALIEN_LINKAGE_SPACE_START;
-#endif
-#if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_LINUX
-    tls[THREAD_MSAN_XOR_CONSTANT_SLOT] = 0x500000000000;
-#endif
-#ifdef LAYOUT_OF_FUNCTION
-    tls[THREAD_FUNCTION_LAYOUT_SLOT] = LAYOUT_OF_FUNCTION << 32;
-#endif
-#ifdef THREAD_TEXT_CARD_MARKS_SLOT
-    extern unsigned int* text_page_touched_bits;
-    tls[THREAD_TEXT_SPACE_ADDR_SLOT] = TEXT_SPACE_START;
-    tls[THREAD_TEXT_CARD_COUNT_SLOT] = text_space_size / IMMOBILE_CARD_BYTES;
-    tls[THREAD_TEXT_CARD_MARKS_SLOT] = (lispobj)text_page_touched_bits;
-#endif
 
     th->os_address = spaces;
     th->control_stack_start = (lispobj*)aligned_spaces;
@@ -984,7 +993,7 @@ alloc_thread_struct(void* spaces) {
         (lispobj*)((char*)th->control_stack_start+thread_control_stack_size);
     th->control_stack_end = th->binding_stack_start;
 
-    if (zeroize_stack) {
+    if (is_recycled) {
 #if GENCGC_IS_PRECISE
     /* Clear the entire control stack. Without this I was able to induce a GC failure
      * in a test which hammered on thread creation for hours. The control stack is
@@ -1011,6 +1020,8 @@ alloc_thread_struct(void* spaces) {
     }
 
     th->state_word.control_stack_guard_page_protected = 1;
+    th->state_word.alien_stack_guard_page_protected = 1;
+    th->state_word.binding_stack_guard_page_protected = 1;
     th->alien_stack_start=
         (lispobj*)((char*)th->binding_stack_start+BINDING_STACK_SIZE);
     set_binding_stack_pointer(th,th->binding_stack_start);
@@ -1023,9 +1034,10 @@ alloc_thread_struct(void* spaces) {
     th->profile_data = (uword_t*)(alloc_profiling ? alloc_profile_buffer : 0);
 
     struct extra_thread_data *extra_data = thread_extra_data(th);
+    // thread_interrupt_data(th) gets cleared by this memset
     memset(extra_data, 0, sizeof *extra_data);
 
-#if defined LISP_FEATURE_SB_THREAD && !defined LISP_FEATURE_SB_SAFEPOINT
+#if THREADS_USING_GCSIGNAL
     os_sem_init(&extra_data->state_sem, 1);
     os_sem_init(&extra_data->state_not_running_sem, 0);
     os_sem_init(&extra_data->state_not_stopped_sem, 0);
@@ -1033,7 +1045,6 @@ alloc_thread_struct(void* spaces) {
 #if defined LISP_FEATURE_UNIX && defined LISP_FEATURE_SB_THREAD
     os_sem_init(&extra_data->sprof_sem, 0);
 #endif
-    extra_data->sprof_lock = 0;
     th->sprof_data = 0;
 
     th->state_word.state = STATE_RUNNING;
@@ -1087,11 +1098,6 @@ alloc_thread_struct(void* spaces) {
     access_control_frame_pointer(th)=0;
 #endif
 
-    thread_interrupt_data(th).pending_handler = 0;
-    thread_interrupt_data(th).gc_blocked_deferrables = 0;
-#if HAVE_ALLOCATION_TRAP_CONTEXT
-    thread_interrupt_data(th).allocation_trap_context = 0;
-#endif
 #if defined LISP_FEATURE_PPC64
     /* Storing a 0 into code coverage mark bytes or GC card mark bytes
      * can be done from the low byte of the thread base register.
@@ -1120,20 +1126,21 @@ alloc_thread_struct(void* spaces) {
     thread_extra_data(th)->synchronous_io_handle_and_flag = 0;
 #endif
     th->stepping = 0;
-    th->card_table = (lispobj)gc_card_mark;
     return th;
 }
 #ifdef LISP_FEATURE_SB_THREAD
 #ifdef LISP_FEATURE_WIN32
-uword_t create_thread(struct thread* th)
+uword_t create_lisp_thread(struct thread* th)
 {
     unsigned int tid;
     struct extra_thread_data *data = thread_extra_data(th);
     data->blocked_signal_set = deferrable_sigset;
     // It's somewhat customary in the win32 API to start threads as suspended.
+    // Don't use STACK_SIZE_PARAM_IS_A_RESERVATION because
+    // dx-allocation accesses the stack non-linearly.
     th->os_thread =
       _beginthreadex(NULL, thread_control_stack_size, new_thread_trampoline, th,
-                     CREATE_SUSPENDED | STACK_SIZE_PARAM_IS_A_RESERVATION, &tid);
+                     CREATE_SUSPENDED, &tid);
     bool success = th->os_thread != 0;
     if (success) {
         th->os_kernel_tid = tid;
@@ -1141,6 +1148,15 @@ uword_t create_thread(struct thread* th)
     }
     return success;
 }
+#else
+#ifdef LISP_FEATURE_OS_THREAD_STACK
+# define START_ROUTINE new_thread_trampoline_switch_stack
+#else
+# define START_ROUTINE new_thread_trampoline
+#endif
+int create_lisp_thread(pthread_t *th, pthread_attr_t *attr, void *arg) {
+    return pthread_create(th, attr, START_ROUTINE, arg);
+}
 #endif
 
 int try_acquire_gc_lock() { return TryEnterCriticalSection(&in_gc_lock); }
@@ -1201,7 +1217,7 @@ int sb_thread_kill (pthread_t thread, int sig) {
 /*
  * (With SB-SAFEPOINT, see the definitions in safepoint.c instead.)
  */
-#if !defined LISP_FEATURE_SB_SAFEPOINT && !defined STANDALONE_LDB
+#if HAVE_GC_STW_SIGNAL && !defined STANDALONE_LDB
 
 /* To avoid deadlocks when gc stops the world all clients of each
  * mutex must enable or disable SIG_STOP_FOR_GC for the duration of
@@ -1337,67 +1353,6 @@ thread_yield()
 #endif
 }
 
-#ifdef LISP_FEATURE_SB_SAFEPOINT
-/* If the thread id given does not belong to a running thread (it has
- * exited or never even existed) pthread_kill _may_ fail with ESRCH,
- * but it is also allowed to just segfault, see
- * <http://udrepper.livejournal.com/16844.html>.
- *
- * Relying on thread ids can easily backfire since ids are recycled
- * (NPTL recycles them extremely fast) so a signal can be sent to
- * another process if the one it was sent to exited.
- *
- * For these reasons, we must make sure that the thread is still alive
- * when the pthread_kill is called and return if the thread is
- * exiting.
- *
- * Note (DFL, 2011-06-22): At the time of writing, this function is only
- * used for INTERRUPT-THREAD, hence the wake_thread special-case for
- * Windows is OK. */
-void wake_thread(struct thread_instance* lispthread)
-{
-#ifdef LISP_FEATURE_WIN32
-        /* META: why is this comment about safepoint builds mentioning
-         * gc_stop_the_world() ? Never the twain shall meet. */
-
-        /* Kludge (on safepoint builds): At the moment, this isn't just
-         * an optimization; rather it masks the fact that
-         * gc_stop_the_world() grabs the all_threads mutex without
-         * releasing it, and since we're not using recursive pthread
-         * mutexes, the pthread_mutex_lock() around the all_threads loop
-         * would go wrong.  Why are we running interruptions while
-         * stopping the world though?  Test case is (:ASYNC-UNWIND
-         * :SPECIALS), especially with s/10/100/ in both loops. */
-
-        /* Frequent special case: resignalling to self.  The idea is
-         * that leave_region safepoint will acknowledge the signal, so
-         * there is no need to take locks, roll thread to safepoint
-         * etc. */
-        struct thread* thread = (void*)lispthread->uw_primitive_thread;
-        if (thread == get_sb_vm_thread()) {
-            sb_pthr_kill(thread, 1); // can't fail
-            check_pending_thruptions(NULL);
-            return;
-        }
-        // block_deferrables + mutex_lock looks very unnecessary here,
-        // but without them, make-target-contrib hangs in bsd-sockets.
-        sigset_t oldset;
-        block_deferrable_signals(&oldset);
-        mutex_acquire(&all_threads_lock);
-        sb_pthr_kill(thread, 1); // can't fail
-# ifdef LISP_FEATURE_SB_SAFEPOINT
-        wake_thread_impl(lispthread);
-# endif
-        mutex_release(&all_threads_lock);
-        thread_sigmask(SIG_SETMASK,&oldset,0);
-#elif defined LISP_FEATURE_SB_SAFEPOINT
-    wake_thread_impl(lispthread);
-#else
-    pthread_kill(lispthread->uw_os_thread, SIGURG);
-#endif
-}
-#endif
-
 #ifdef LISP_FEATURE_ULTRAFUTEX
 extern int futex_wake(int *lock_word, int n);
 void lispmutex_wake_waiter()
